@@ -1,31 +1,28 @@
 import asyncio
-from asyncio import subprocess
 import datetime
 import gzip
 import json
 import logging
 import multiprocessing
 import sys
+from functools import partial
 from pathlib import Path
 from typing import Any
 
 import httpx
 import pandas as pd
 import requests
-
 import tqdm
+from tqdm.contrib.concurrent import process_map
 from tqdm.contrib.logging import logging_redirect_tqdm
 
-from gpas import data_dir, misc, validation
+from gpas import misc
 from gpas.misc import (
     DEFAULT_ENVIRONMENT,
     ENDPOINTS,
     ENVIRONMENTS,
     FILE_TYPES,
     GOOD_STATUSES,
-    AuthenticationError,
-    DecontaminationError,
-    SubmissionError,
 )
 from gpas.validation import build_validation_message, validate
 
@@ -69,7 +66,7 @@ def fetch_user_details(access_token, environment: ENVIRONMENTS):
     except requests.exceptions.HTTPError as e:
         status_code = e.response.status_code
         if status_code == 401:
-            raise AuthenticationError(
+            raise misc.AuthenticationError(
                 f"Authentication failed, check access token and environment"
             ) from None
         else:
@@ -139,7 +136,7 @@ async def fetch_status_single_async(
     client, guid, url, headers, warn=False, n_retries=5
 ):
     r = await client.get(url=url, headers=headers)
-    if r.status_code == httpx.codes.ok:
+    if r.status_code == httpx.codes.OK:
         r_json = r.json()[0]
         status = r_json.get("status")
         result = dict(sample=guid, status=status)
@@ -225,7 +222,7 @@ async def download_single_async(
     prefix = name if name else guid
     r = await client.get(url=url, headers=headers)
     Path(out_dir).mkdir(parents=False, exist_ok=True)
-    if r.status_code == httpx.codes.ok:
+    if r.status_code == httpx.codes.OK:
         logging.debug(
             Path(out_dir) / Path(f"{prefix}.{file_types_extensions[file_type]}")
         )
@@ -339,7 +336,6 @@ class Sample:
         self.decontamination_ref_path = self.get_decontamination_ref_path()
         self.working_dir = Path(working_dir)
         self.working_dir.mkdir(parents=False, exist_ok=True)
-        self.uploaded = False
         self.guid = None
         self.mapping_path = None
         self.samtools_path = samtools_path
@@ -355,22 +351,7 @@ class Sample:
         if self.specimen_organism == "SARS-CoV-2":
             cmd = self._get_riak_cmd()
         else:
-            raise DecontaminationError("Invalid organism")
-
-        # before_msg = {
-        #     "progress": {
-        #         "action": "decontamination",
-        #         "status": "started",
-        #         "sample": self.sample_name,
-        #     }
-        # }
-        # after_msg = {
-        #     "progress": {
-        #         "action": "decontamination",
-        #         "status": "finished",
-        #         "sample": self.sample_name,
-        #     }
-        # }
+            raise misc.DecontaminationError("Invalid organism")
 
         command = misc.LoggedShellCommand(
             name=self.sample_name,
@@ -382,33 +363,18 @@ class Sample:
     def _get_convert_bam_cmd(self, paired=False) -> str:
         prefix = Path(self.working_dir) / Path(self.sample_name)
         if not self.paired:
-            cmd = f"{self.samtools_path} fastq -0 {prefix.with_suffix('.fastq.gz')} {self.bam}"
+            cmd = f'{self.samtools_path} fastq -0 "{prefix.with_suffix(".fastq.gz")}" "{self.bam}"'
             self.fastq = self.working_dir / Path(self.sample_name + ".fastq.gz")
         else:
             cmd = (
-                f"{self.samtools_path} sort {self.bam} |"
+                f'{self.samtools_path} sort "{self.bam}" |'
                 f" {self.samtools_path} fastq -N"
-                f" -1 {prefix.parent / (prefix.name + '_1.fastq.gz')}"
-                f" -2 {prefix.parent / (prefix.name + '_2.fastq.gz')}"
+                f' -1 "{prefix.parent / (prefix.name + "_1.fastq.gz")}"'
+                f' -2 "{prefix.parent / (prefix.name + "_2.fastq.gz")}"'
             )
             self.fastq1 = self.working_dir / Path(self.sample_name + "_1.fastq.gz")
             self.fastq2 = self.working_dir / Path(self.sample_name + "_2.fastq.gz")
 
-        # before_msg = {
-        #     "progress": {
-        #         "action": action,
-        #         "status": "started",
-        #         "sample_name": self.sample_name,
-        #     }
-        # }
-        # after_msg = {
-        #     "progress": {
-        #         "action": action,
-        #         "status": "finished",
-        #         "sample_name": self.sample_name,
-        #     }
-        # }
-
         command = misc.LoggedShellCommand(
             name=self.sample_name,
             action="bam_conversion",
@@ -421,17 +387,17 @@ class Sample:
         if not self.fastq2:
             cmd = (
                 f"{self.decontaminator_path} --tech ont --enumerate_names"
-                f" --ref_fasta {self.decontamination_ref_path}"
-                f" --reads1 {self.fastq}"
-                f" --outprefix {self.working_dir / self.sample_name}"
+                f' --ref_fasta "{self.decontamination_ref_path}"'
+                f' --reads1 "{self.fastq}"'
+                f' --outprefix "{self.working_dir / self.sample_name}"'
             )
         else:
             cmd = (
                 f"{self.decontaminator_path} --tech illumina --enumerate_names"
-                f" --ref_fasta {self.decontamination_ref_path}"
-                f" --reads1 {self.fastq1}"
-                f" --reads2 {self.fastq2}"
-                f" --outprefix {self.working_dir / self.sample_name}"
+                f' --ref_fasta "{self.decontamination_ref_path}"'
+                f' --reads1 "{self.fastq1}"'
+                f' --reads2 "{self.fastq2}"'
+                f' --outprefix "{self.working_dir / self.sample_name}"'
             )
         self.clean_fastq = (
             self.working_dir / Path(self.sample_name + ".reads.fastq.gz")
@@ -466,34 +432,6 @@ class Sample:
             "gpas_sample_name": self.guid,
         }
 
-    def _upload_reads(self, batch_url, headers, json_messages):
-        """Upload an unpaired FASTQ file to the Organisation's input bucket in OCI"""
-        url_prefix = batch_url + self.guid
-        if not self.uploaded:
-            if json_messages:
-                misc.print_progress_message_json(
-                    action="upload", status="started", sample=self.sample_name
-                )
-            if not self.paired:
-                with open(self.clean_fastq, "rb") as fh:
-                    r = requests.put(
-                        url=f"{url_prefix}.reads.fastq.gz", data=fh, headers=headers
-                    )
-            else:
-                with open(self.clean_fastq1, "rb") as fh:
-                    r = requests.put(
-                        f"{url_prefix}.reads_1.fastq.gz", data=fh, headers=headers
-                    )
-                with open(self.clean_fastq2, "rb") as fh:
-                    r = requests.put(
-                        f"{url_prefix}.reads_2.fastq.gz", data=fh, headers=headers
-                    )
-            if json_messages:
-                misc.print_progress_message_json(
-                    action="upload", status="started", sample=self.sample_name
-                )
-        self.uploaded = True
-
 
 class Batch:
     """
@@ -515,9 +453,7 @@ class Batch:
         self.environment = environment
         self.working_dir = working_dir
         self.out_dir = out_dir
-        self.processes = (
-            processes if processes else int(multiprocessing.cpu_count() / 2)
-        )
+        self.processes = processes if processes else int(multiprocessing.cpu_count())
         self.json = {"validation": "", "decontamination": "", "submission": ""}
         self.json_messages = json_messages
         self.samtools_path = misc.get_binary_path("samtools")
@@ -563,9 +499,9 @@ class Batch:
         )
 
         self._number_runs()
-
+        self.uploaded = False
         if self.json_messages:
-            print(json.dumps(self.validation_json, indent=4))
+            misc.print_json(self.validation_json)
 
     def _get_convert_bam_cmds(self) -> list[misc.LoggedShellCommand]:
         return [s._get_convert_bam_cmd() for s in self.samples]
@@ -597,10 +533,8 @@ class Batch:
         samples_decontamination_stats = {
             s: parse_decontamination_stats(r.stdout) for s, r in samples_runs.items()
         }
-        for sample in self.samples:
-            sample.decontamination_stats = samples_decontamination_stats.get(
-                sample.sample_name
-            )
+        for s in self.samples:
+            s.decontamination_stats = samples_decontamination_stats.get(s.sample_name)
 
     def _hash_fastqs(self):
         if not self.paired:
@@ -724,23 +658,66 @@ class Batch:
         self.bucket = self.par.split("/")[-3]
         self.batch_url = self.par + self.batch_guid + "/"
 
-    def _upload_samples(self):
-        if self.json_messages:  # Avoid tqdm completely for PyInstaller's sake
+    def _get_uploads(self) -> list[misc.SampleUpload]:
+        """Return local file path and destination URL for the file to be uploaded"""
+        uploads = []
+        for s in self.samples:
+            url_prefix = self.batch_url + s.guid
+            if not s.paired:
+                uploads.append(
+                    misc.SampleUpload(
+                        name=s.sample_name,
+                        path1=s.clean_fastq,
+                        path2=None,
+                        url1=f"{url_prefix}.reads.fastq.gz",
+                        url2=None,
+                    )
+                )
+            else:
+                uploads.append(
+                    misc.SampleUpload(
+                        name=s.sample_name,
+                        path1=s.clean_fastq1,
+                        path2=s.clean_fastq2,
+                        url1=f"{url_prefix}.reads_1.fastq.gz",
+                        url2=f"{url_prefix}.reads_2.fastq.gz",
+                    )
+                )
+        return uploads
+
+    def _upload_samples(self) -> None:
+        if self.json_messages:
             misc.print_progress_message_json(action="upload", status="started")
-            for s in self.samples:
-                s._upload_reads(self.batch_url, self.headers, self.json_messages)
-                logging.info(f"Finished uploading {s.sample_name} ({s.guid})")
-            misc.print_progress_message_json(action="upload", status="finished")
+        uploads = self._get_uploads()
+        if self.processes == 1:
+            for upload in uploads:
+                misc.upload_sample(
+                    upload=upload,
+                    headers=self.headers,
+                    json_messages=self.json_messages,
+                )
         else:
-            for s in tqdm.tqdm(
-                self.samples,
-                desc=f"Uploading",
+            process_map(
+                partial(
+                    misc.upload_sample,
+                    headers=self.headers,
+                    json_messages=self.json_messages,
+                ),
+                uploads,
+                max_workers=10,
+                desc=f"Uploading {len(uploads)} sample(s) (10 connections)",
                 bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                 leave=False,
-            ):
-                s._upload_reads(self.batch_url, self.headers, self.json_messages)
+            )
+        self.uploaded = True
+        if self.json_messages:
+            misc.print_progress_message_json(action="upload", status="finished")
+        else:
+            logging.info(f"Finished uploading {len(uploads)} sample(s)")
 
     def _finalise_submission(self):
+        if not self.uploaded:
+            raise RuntimeError("Files not uploaded")
         endpoint = (
             ENDPOINTS[self.environment.value]["HOST"]
             + ENDPOINTS[self.environment.value]["ORDS_PATH"]
@@ -750,7 +727,7 @@ class Batch:
         r.raise_for_status()
         logging.debug(f"POSTing JSON {r.text=}")
         if r.json().get("status") != "success":
-            raise SubmissionError(r.json().get("errorMsg"))
+            raise misc.SubmissionError(r.json().get("errorMsg"))
         url = self.par + self.batch_guid + "/upload_done.txt"  # Finalisation mark
         r = requests.put(url=url, headers=self.headers)
         r.raise_for_status()
@@ -763,16 +740,13 @@ class Batch:
             }
         }
         if self.json_messages:
-            print(json.dumps(success_message, indent=4))
+            misc.print_json(success_message)
 
     def _submit(self):
-        # try:
         self._fetch_par()
         self._upload_samples()
         self._prepare_submission()
         self._finalise_submission()
-        # except Exception as e:
-        #     raise SubmissionError(e) from None
 
     def _prepare_submission(self):
         """Prepare the JSON payload for the GPAS Upload app
