@@ -1,23 +1,29 @@
 import hashlib
 import json
 import logging
-from multiprocessing.dummy import Pool
 import os
 import shutil
 import subprocess
 import sys
 import traceback
+from multiprocessing.pool import ThreadPool
 from dataclasses import dataclass
-from functools import partial
 from enum import Enum
+from functools import partial
 from pathlib import Path
 
 import pandas as pd
+import requests
 import tqdm
+from tenacity import (
+    retry,
+    retry_if_exception_type,
+    wait_exponential,
+    stop_after_attempt,
+)
 from tqdm.contrib.logging import logging_redirect_tqdm
 
-import gpas
-from gpas import validation
+from gpas import data_dir, validation
 
 
 FORMATS = Enum("Formats", dict(table="table", csv="csv", json="json"))
@@ -72,6 +78,15 @@ class LoggedShellCommand:
     cmd: str
 
 
+@dataclass
+class SampleUpload:
+    name: str
+    path1: Path
+    path2: str | None
+    url1: str
+    url2: str | None
+
+
 def get_value_traceback(e: Exception) -> tuple[str, str, list]:
     e_type, e_value, e_traceback = sys.exc_info()
     e_t = str(e_type)
@@ -119,24 +134,13 @@ def print_progress_message_json(action: str, status: str, sample: str = ""):
     }
     if sample:
         message["progress"]["sample"] = sample
-    print(json.dumps(message, indent=4), flush=True)
-
-
-# def print_progress_message(LoggedShellCommand, status: str):
-#     message = {
-#         "progress": {
-#             "action": action,
-#             "status": status,
-#         }
-#     }
-#     print(json.dumps(message, indent=4), flush=True)
+    print_json(message)
 
 
 def run_logged(
     command: LoggedShellCommand, json_messages: bool = False
 ) -> subprocess.CompletedProcess:
     if json_messages:
-        # logging.basicConfig(format="%(message)s", level=logging.INFO)
         print_progress_message_json(
             action=command.action, status="started", sample=command.name
         )
@@ -147,10 +151,6 @@ def run_logged(
         print_progress_message_json(
             action=command.action, status="finished", sample=command.name
         )
-    # else:
-    #     with logging_redirect_tqdm():
-    #         logging.info(f"Finished {command.action} for {command.name}")
-
     return process
 
 
@@ -168,12 +168,11 @@ def run_parallel_logged(
         for c in commands:
             results[c.name] = run_logged(command=c, json_messages=json_messages)
             logging.debug(f"{c.cmd=}")
-            # logging.info(f"Finished {participle.lower()} {len(cmds)} sample(s)")
     else:
         names = [c.name for c in commands]
         cmds = [c.cmd for c in commands]
         logging.debug(f"Started {participle.lower()} {len(cmds)} sample(s) \n{cmds=}")
-        with Pool(processes) as pool:
+        with ThreadPool(10) as pool:
             results = {
                 n: c
                 for n, c in zip(
@@ -194,7 +193,7 @@ def run_parallel_logged(
         print_progress_message_json(action=commands[0].action, status="finished")
     else:
         with logging_redirect_tqdm():
-            logging.info(f"Finished {participle.lower()} {len(cmds)} sample(s)")
+            logging.info(f"Finished {participle.lower()} {len(commands)} sample(s)")
     return results
 
 
@@ -242,20 +241,15 @@ def get_binary_path(filename: str) -> str:
     env_var = f"GPAS_{filename.upper()}_PATH"
     if os.getenv(env_var) and Path(os.environ[env_var]).exists():
         path = Path(os.environ[env_var]).resolve()
+        logging.debug(f"get_binary_path(): Environment variable mode {path=}")
     elif hasattr(sys, "_MEIPASS"):  # PyInstaller onefile
-        path = Path(sys.executable).parent
-        print("Using", str(path), "!")
-    elif (Path(__file__).parents[0] / filename).exists():
-        path = (Path(__file__).parents[0] / filename).resolve()
-    elif (Path(__file__).parents[1] / filename).exists():
-        path = (Path(__file__).parents[1] / filename).resolve()
-    elif (Path(__file__).parents[2] / filename).exists():
-        path = (Path(__file__).parents[2] / filename).resolve()
+        path = (Path(sys.executable).parent / filename).resolve()
+        logging.debug(f"get_binary_path(): PyInstaller mode {path=}")
     elif shutil.which(filename):  # $PATH
         path = Path(shutil.which(filename)).resolve()
+        logging.debug(f"get_binary_path(): $PATH mode {path=}")
     else:
         raise FileNotFoundError(f"Could not find {filename} binary")
-    logging.debug(f"{filename=} {path=}")
     return str(path)
 
 
@@ -275,12 +269,12 @@ def get_data_path():
     env_var = "GPAS_DATA_PATH"
     if os.environ.get(env_var) and Path(os.environ[env_var]).exists():
         path = Path(os.environ[env_var]).resolve()
-    elif gpas.data_dir.exists():
-        path = gpas.data_dir
+    elif data_dir.exists():
+        path = data_dir
     elif (Path(__file__).parents[1] / "data").exists():
         path = Path(__file__).parents[1] / "data"
     else:
-        print(f"{__name__=} {__file__=} {gpas.data_dir=}")
+        logging.debug(f"{__name__=} {__file__=} {data_dir=}")
         raise FileNotFoundError(f"Could not find data directory")
     return path.resolve()
 
@@ -289,3 +283,27 @@ def get_reference_path(organism):
     prefix = get_data_path() / "refs"
     organisms_paths = {"SARS-CoV-2": "MN908947_no_polyA.fasta"}
     return Path(prefix / organisms_paths[organism]).resolve()
+
+
+@retry(
+    retry=retry_if_exception_type(requests.exceptions.ConnectionError),
+    wait=wait_exponential(multiplier=1, min=1, max=5),
+    stop=stop_after_attempt(5),
+)
+def upload_sample(upload: SampleUpload, headers: dict, json_messages: bool) -> None:
+    if json_messages:
+        print_progress_message_json(
+            action="upload", status="started", sample=upload.name
+        )
+    with open(upload.path1, "rb") as fh:
+        r = requests.put(url=upload.url1, data=fh, headers=headers)
+        r.raise_for_status()
+    if upload.path2 and upload.url2:
+        with open(upload.path2, "rb") as fh:
+            r = requests.put(url=upload.url2, data=fh, headers=headers)
+            r.raise_for_status()
+    logging.debug(f"Uploaded sample {upload.name}")
+    if json_messages:
+        print_progress_message_json(
+            action="upload", status="finished", sample=upload.name
+        )
